#!/bin/sh
#
# copynilfs - backup tool of NILFS partitions
# Copyright (C) 2006 Nippon Telegraph and Telephone Corporation.
# This file is part of NILFS and is licensed under the GPL.
#
# copynilfs.sh,v 1.16 2007/02/11 13:36:12 ryusuke Exp
#
# Written by Ryusuke Konishi <ryusuke@osrg.net>
#
cmd=`basename $0`

inspect=/sbin/inspect
sharedir=/usr/share/nilfsprogs
filterdir=$sharedir/copynilfs/filter.d
tmpdir=/tmp

show_help() {
    echo "Usage: $cmd [options..] <source device> [<host name>:]<destination device>"
    echo "Options:"
#    echo "  -a                append mode"
    echo "  -h                show this help"
    echo "  -f <cp-list>      read first columns of the <cp-list> file as"
    echo "                    selected checkpoints (list of block # of segment)"
    echo "  -F <filter>       specify a filter command that selects checkpoints"
    echo "  -t <nr-blocks>    set threshold value of dirty data blocks"
    echo "                    This value affects granularity of the segments"
    echo "                    to be written on the new device"
    echo "  -X                Do not preserve hard links"
    echo "  -q                quiet mode"
    echo "  -l <user name>    remote user who has root privilleges"
    echo "Newfs options:"
    echo "  -b <block-size>   specify blocksize of the destination device"
    echo "  -B <integer>      specify number of blocks per segment"
    echo "  -L <label>        specify label of new device"
}

function parse_filter() {
    local fcmd=`expr "$1" : "\(\S\+\)"`
    local fargs=`expr "$1" : "\S\+\s\+\(.*\)$"`

    if [ -x "$fcmd" ]; then :
	test `expr substr "$fcmd" 1 1` = '/' || fcmd=`pwd`/"$fcmd"
    elif [ -x "$filterdir/$fcmd" ]; then
	fcmd="$filterdir/$fcmd"
    else
	local fpath=`which $fcmd`
	if [ -z "$fpath" -o ! -x "$fpath" ]; then
	    echo "Not executable filter: $fcmd" 1>&2; exit 1
	fi
    fi
    echo "$fcmd $fargs"
}

quiet=0
copy_devices=1
copy_hardlinks=1
cpfilter=""
remote_user="${USER:-root}"
mode=0  # client mode
unset block_per_seg block_size vlabel threshold

while getopts ahf:F:qb:B:L:t:Xl:D o; do
    case $o in
	h)
	    show_help 1>&2
	    exit 0 ;;
	a)
	    echo "Not supported yet"
	    exit 1 ;;
	f)
	    cplist="$OPTARG" ;;
	F)
	    fline=`parse_filter "$OPTARG"` || exit $?
	    cpfilter=${cpfilter:+"$cpfilter | "}"$fline"' $1' ;;
	q)
	    quiet=1 ;;
	b)
	    block_size="$OPTARG" ;;
	B)
	    block_per_seg="$OPTARG" ;;
	L)
	    vlabel="$OPTARG" ;;
	t)
	    threshold="$OPTARG" ;;
	X)
	    copy_hardlinks=0 ;;
	l)
	    remote_user="$OPTARG" ;;
	D)
	    mode=1 ;;  # destination server mode
	*)
	    show_help 1>&2
	    exit 1 ;;
    esac
done
shift `expr $OPTIND - 1`

# Sub-routines
function sysfs_nilfs_dir() {
    local bdev=`basename $1`
    local hdd=`expr $bdev : "\([[:alnum:]]*[[:alpha:]]\)[0-9]\+$"`

    if [ -z "$hdd" -o ! -d "/sys/block/$hdd/$bdev/nilfs" ]; then
	echo "/sys/block/$bdev/nilfs"
    else
	echo "/sys/block/$hdd/$bdev/nilfs"
    fi
}

if [ $mode -eq 0 ]; then
    if [ -z "$1" ]; then
	show_help 1>&2; exit 1
    elif [ ! -b "$1" ]; then
	echo "Not device: $1" 1>&2; exit 1
    elif [ "$1" == "$2" ]; then
	echo "Identical devices" 1>&2; exit 1
    elif [ ! -r "$1" ]; then
	echo "Device not readable: $1" 1>&2; exit 1
    elif [ ! -x $inspect ]; then
	echo "Please install inspect" 1>&2; exit 1;
    elif [ -n "$cplist" -a ! -f "$cplist" ]; then
	echo "Cannot find checkpoint list file: $cplist"; exit 1
    fi
    sdev="$1"; shift 1
else
    export PATH=/usr/sbin:/usr/bin:/sbin:/bin
fi

dhost=`expr "$1" : '\([^:]\+\):[[:print:]]\+$'`
ddev=`expr "$1" : '[^:]\+:\([[:print:]]\+\)$'`
if [ -z "$ddev" ]; then ddev="$1"; fi
#echo "dhost=$dhost" 1>&2
#echo "ddev=$ddev" 1>&2

tdir=$tmpdir/$cmd.$PPID
sdir=$tdir/src.d
ddir=$tdir/dst.d
dpipe=$tdir/dst.p
err=0

# Common checks
if [ -z "$ddev" ]; then
    show_help 1>&2; exit 1
elif [ ! -x `which rsync` ]; then
    echo "Cannot find rsync command" 1>&2; exit 1;
fi

function getcp() {
    (echo "listcp"; echo "quit") | $inspect -e $1 | sed -ne '/MajorCP/{s/^nilfs> //;p}'
}

# Get superblock information
function getsb() {
    echo "quit" | $inspect -e $1 |
    ( while read h v o; do
	case "$h" in
	    s_ctime) 
		ctime="$v" ;;
	    block_size)
		test "$v" -eq 4096 || : ${block_size:="$v"} ;;
	    s_blocks_per_segment)
		test "$v" -eq 1024 || : ${block_per_seg:="$v"} ;;
	    s_volume_name)
		test -z "$v" || : ${vlabel:="$v"} ;;
#	s_uuid)
#	    : ${uuid:="$v"} ;;
	esac
    done ;
    echo "${ctime:+-P $ctime}${block_size:+ -b $block_size}${block_per_seg:+ -B $block_per_seg}${vlabel:+ -L $vlabel}" )
}

function read_cplist() {
    while read ncp o <&3; do
	ncp=`expr "$ncp" : '0*\([0-9]\+\)$'`
	if [ -n "$ncp" ] && [ $ncp -gt $1 ]; then return 0; fi
    done
    return 1
}

# Common routines
function mounted() { test `df -t nilfs $1 | wc -l` -ne 1; }

function abort() { 
    if [ -n "$1" ]; then echo "$1" 1>&2; else echo "Aborting.." 1>&2; fi
    if mounted $sdir; then umount $sdir; fi
    wait
    rm -rf $tdir; exit 1
}

# Destination Server
function dserv() {
    local ddev="$1"
    local sysfs_dir="" err=1 mnt=0 rargs=""

    if [ ! -b $1 ]; then
	echo "Not device: $1" 1>&2
    elif [ ! -r $1 ]; then
	echo "Device not readable: $1" 1>&2
    elif [ ! -w $1 ]; then
	echo "Device not writable: $1" 1>&2
    elif ! mkdir -p $ddir 1>&2; then
	echo "Cannot make $ddir" 1>&2
    else
	err=0
    fi
    echo $err

    while [ $err -eq 0 ] && read cmd args; do
#	echo "dserv: ddir=$ddir ddev=$1" 1>&2
#	echo "dserv: $cmd $args" 1>&2
	rargs=""

	case $cmd in
	    test)
		echo "Test" 1>&2
		err=0 ;;
	    mkfs)
		mkfs -t nilfs $args $ddev 1>&2
		err=$? ;;
	    mount)
		local mntdir=`awk '$1 == "'$1'"{print $2}' /etc/mtab`

		err=1
		if [ -n "$mntdir" ]; then
		    echo "Device $1 is mounted on $mntdir" 1>&2
		elif ! mount -t nilfs -o passive=on $ddev $ddir 1>&2 ; then
		    echo "mount failed." 1>&2
		else
		    mnt=1
		    sysfs_dir=`sysfs_nilfs_dir $1`
		    if [ ! -w $sysfs_dir/sc_ctime ]; then
			echo "Cannot find sc_ctime." 1>&2
		    elif [ ! -w $sysfs_dir/sync ]; then
			echo "Cannot find sync file." 1>&2
		    else 
			if [ -n "$threshold" -a -w $sysfs_dir/sc_threshold ]; then
			    echo "$threshold" > $sysfs_dir/sc_threshold
			fi
			err=0; rargs="$ddir"
		    fi
		fi ;;
	    settime)
		echo $args > $sysfs_dir/sc_ctime
		err=$? ;;
	    sketch)
		if [ -f $ddir/.sketch ]; then err=0; else err=1; fi ;;
	    sync)
		echo 1 > $sysfs_dir/sync
		err=$? ;;
	    quit)
		err=0; break ;;
	    *)
		echo "Unknown command: $cmd" 1>&2; err=1 ;;
	esac
	sleep 0
	echo $err $rargs
    done
    if [ $mnt -ne 0 ]; then umount $ddir 1>&2; fi
}

# Main part
function copy_loop() {
    local sk=.sketch ncp="" pcp=1
    local b c t o
    local xopt="${dhost:+-e ssh -z}"
    local ddir2="${dhost:+${remote_user:+${remote_user}@}${dhost}:}$ddir"
    local cp="${dhost:+s}cp ${dhost:+-q}"

    while read b c t o; do
	test $b -ne 1 || continue
	if [ -n "$cplist" ]; then
	    test -n "$ncp" || read_cplist $pcp || break
	    while [ $ncp -ne $b ]; do
		test $ncp -gt $b || read_cplist $b || break 2
		continue 2
	    done
	    pcp=$ncp; ncp=""
	fi
	test $quiet -ne 0 || echo "segblk=$b" 1>&2
	if mount -t nilfs -r -o cp=$b $1 $sdir 1>&2 ; then
	    if [ -n "$sk" -o ! -f "$sdir/$sk" ]; then
		sk=`cd $sdir; find -maxdepth 1 -mount -inum 10 -print | sed -e 's|^\./||'`
	    fi

	    echo settime $t
	    read err args < $dpipe || return 1
	    err=1
	    if rsync -ax $xopt $3 --inplace ${sk:+--exclude="$sk"} --del $sdir/ $ddir2/ 1>&2; then
		echo sketch
		read err args < $dpipe || return 1
		if [ "$err" == 0 ]; then
		    # echo "copying sketch to $ddir2/$sk with $cp (cp=$b)" 1>&2
		    if [ -f "$tdir/sketch.$b" ]; then
			$cp -p "$tdir/sketch.$b" "$ddir2/$sk" 1>&2; err=$?
		    elif [ -s "$sdir/$sk" ]; then
			$cp -p "$sdir/$sk" "$ddir2/$sk" 1>&2; err=$?
		    fi
		fi
		if [ "$err" == 0 ]; then
		    echo sync
		    read err args < $dpipe || return 1
		fi
	    fi
	    umount $sdir 1>&2
	    if [ $err -ne 0 ]; then
		echo "Failed to write snapshot(cp=$b)" 1>&2; return $err
	    fi
	else
	    echo "Failed to mount $1(cp=$b) on $sdir. skipped" 1>&2
	fi
    done
}

function copy_nilfs() {
    local opt=""

    test $copy_hardlinks -eq 0 || opt="-H"
    test $copy_devices -eq 0 -o $UID -ne 0 || opt="$opt -D"

    test $quiet -ne 0 || echo "Copying checkpoints:" 1>&2
    if [ -n "$cplist" ]; then
	copy_loop $1 $2 "$opt" 3< "$cplist"
    else
	copy_loop $1 $2 "$opt"
    fi
    err=$?
    test $err -eq 0 || return $err
    test $quiet -ne 0 || echo "Done." 1>&2
}

function scli() {
    read err args < $dpipe || abort
    test "$err" == 0 || abort

    echo mkfs "$newfs_opt"
    read err args < $dpipe || abort
    test "$err" == 0 || abort

    echo mount
    read err ddir args < $dpipe || abort
    test "$err" == 0 || abort

    if [ -n "$cpfilter" ]; then
	getcp "$1" | eval '(cd $tdir; '"$cpfilter"')' | copy_nilfs "$1" "$2"
    else
	getcp "$1" | copy_nilfs "$1" "$2"
    fi
    err=$?

    echo quit
}

mkdir -m go-rxw -p $tdir

case $mode in
    0)
	echo "$cmd will overwrite all existing data on ${dhost:+${dhost}:}$ddev." 1>&2
	if ! read -p "Do you proceed [y/N]? " yn || [ "$yn" != "y" -a "$yn" != "Y" ]; then
	    rm -rf $tdir
	    exit 0;
	fi

	newfs_opt=`getsb $sdev`

	mkdir -p $sdir || abort "Cannot make $sdir"

	dssh="${dhost:+ssh ${remote_user:+-l $remote_user} $dhost}"
	mkfifo $dpipe || abort "Cannot make named pipe $dpipe"

	trap abort SIGINT SIGTERM

	scli "$sdev" "$ddev" | $dssh copynilfs -D "$ddev" 1> $dpipe

	trap "" SIGINT SIGTERM
	;;
    1)
	dserv "$ddev"
	;;
esac

wait
rm -rf $tdir

exit $err
